"""
Utilities module.
"""

import os
import numpy as np


# pylint: disable=unused-argument
# noinspection PyUnusedLocal
def array_to_binary(ar, obj=None, force_contiguous=True):
    """Pre-process numpy array for serialization in traittypes.Array."""
    if ar.dtype.kind not in ['u', 'i', 'f']:  # ints and floats
        raise ValueError("unsupported dtype: %s" % ar.dtype)

    if ar.dtype == np.float64:  # WebGL does not support float64
        ar = ar.astype(np.float32)
    elif ar.dtype == np.int64:  # JS does not support int64
        ar = ar.astype(np.int32)

    if force_contiguous and not ar.flags["C_CONTIGUOUS"]:  # make sure it's contiguous
        ar = np.ascontiguousarray(ar)

    return {'buffer': memoryview(ar), 'dtype': str(ar.dtype), 'shape': ar.shape}


# noinspection PyUnusedLocal
def from_json_to_array(value, obj=None):
    """Post-process traittypes.Array after deserialization to numpy array."""
    if value:
        return np.frombuffer(value['buffer'], dtype=value['dtype']).reshape(value['shape'])
    return None


array_serialization = dict(to_json=array_to_binary, from_json=from_json_to_array)


def download(url):
    """Retrieve the file at url, save it locally and return the path."""
    basename = os.path.basename(url)
    if os.path.exists(basename):
        return basename

    print('Downloading: {}'.format(basename))

    # 2/3 compatibility hacks
    try:
        from urllib.request import urlopen
    except ImportError:
        import urllib2
        import contextlib
        urlopen = lambda url_: contextlib.closing(urllib2.urlopen(url_))

    with urlopen(url) as response, open(basename, 'wb') as output:
        output.write(response.read())

    return basename


def check_attribute_range(attribute, color_range):
    if attribute.size == 0 or len(color_range) == 2:
        return color_range
    color_range = [np.min(attribute), np.max(attribute)]
    if color_range[0] == color_range[1]:
        color_range[1] += 1.0
    return color_range


def map_colors(attribute, color_map, color_range=()):
    a_min, a_max = check_attribute_range(attribute, color_range)
    map_array = np.asarray(color_map)
    map_array = map_array.reshape((map_array.size // 4, 4))
    attribute = (attribute - a_min) / (a_max - a_min)  # normalizing attribute for range lookup
    red, green, blue = [np.array(255 * np.interp(attribute, xp=map_array[:, 0], fp=map_array[:, i+1]), dtype=np.int32)
                        for i in range(3)]
    colors = (red << 16) + (green << 8) + blue
    return colors



import unittest
from ..objects import Drawable
from ..k3d import text
import numpy as np


class TestDrawable(unittest.TestCase):
    def setUp(self):
        self.obj = Drawable()

    def test_iteration_returns_object(self):
        for obj in self.obj:
            self.assertTrue(isinstance(obj, Drawable))

    def test_can_add_another_object(self):
        self.obj += Drawable()

        for obj in self.obj:
            self.assertTrue(isinstance(obj, Drawable))

    def test_can_add_many_objects(self):
        obj = Drawable()
        self.obj += obj + obj

        for obj in self.obj:
            self.assertTrue(isinstance(obj, Drawable))


class TestText(unittest.TestCase):
    def test_position_accepts_numpy(self):
        text_ = text('test', [0, 0, 0])
        text_.position = np.arange(3)


if __name__ == '__main__':
    unittest.main()



import urllib.request
import json

response = urllib.request.urlopen('https://raw.githubusercontent.com/Kitware/ParaView/master/ParaViewCore/' +
                                  'ServerManager/Rendering/ColorMaps.json')
data = json.loads(response.read().decode('utf8'))

file = open('paraview_color_maps.py', 'w')

for item in data:
    if 'RGBPoints' in item:
        name = item['Name'].replace(' ', '_').replace('-', '_').replace('(', '').replace(')', '').replace('2', 'two_')\
            .replace(',', '')
        name = name[:1].upper() + name[1:]
        file.write(name + ' = [\n')

        list_ = [item['RGBPoints'][i:i + 4] for i in range(0, len(item['RGBPoints']), 4)]

        for p in list_:
            file.write('        ' + str(p[0]) + ', ' + str(p[1]) + ', ' + str(p[2]) + ', ' + str(p[3]) + ',\n')

        file.write(']\n\n')

file.close()



import ipywidgets as widgets
from traitlets import Unicode, Int, Float, List, Bool, Bytes, Integer
from traitlets import validate, TraitError
from traittypes import Array
from .helpers import array_serialization
from ._version import __version__
import numpy as np


class ListOrArray(List):
    _cast_types = (tuple, np.ndarray)

    def __init__(self, *args, **kwargs):
        self._empty_ok = kwargs.pop('empty_ok', False)
        super(ListOrArray, self).__init__(*args, **kwargs)

    def validate_elements(self, obj, value):
        if self._empty_ok and len(value) == 0:
            return list(value)
        return super(ListOrArray, self).validate_elements(obj, value)


class Drawable(widgets.CoreWidget):
    """
    Base class for drawable objects and groups.
    """

    _model_name = Unicode('ObjectModel').tag(sync=True)
    _model_module = Unicode('k3d').tag(sync=True)
    _model_module_version = Unicode('~' + __version__).tag(sync=True)

    id = Integer().tag(sync=True)

    def __init__(self, **kwargs):
        self.id = id(self)

        super(Drawable, self).__init__(**kwargs)

    def __iter__(self):
        return (self,).__iter__()

    def __add__(self, other):
        return Group(self, other)

    def fetch_data(self, field):
        """Request updating the value of a field modified in browser.

        For data modified in the widget on the browser side, this triggers an asynchronous
        update of the value in the Python kernel.

        Only specific features require this mechanism, e.g. the in-browser editing of voxels.

        Arguments:
            field: `str`. the field name."""
        self.send({'msg_type': 'fetch', 'field': field})

    def _ipython_display_(self, **kwargs):
        """Called when `IPython.display.display` is called on the widget."""
        from IPython.display import display
        import k3d
        plot = k3d.plot()
        plot += self
        return display(plot, **kwargs)


class Group(Drawable):
    """
    An aggregated group of Drawables, itself a Drawable.

    It can be inserted or removed from a Plot including all members.
    """

    __objs = None

    def __init__(self, *args):
        self.__objs = tuple(self.__assert_drawable(drawable) for drawables in args for drawable in drawables)

    def __iter__(self):
        return self.__objs.__iter__()

    def __setattr__(self, key, value):
        """Special method override which allows for setting model matrix for all members of the group."""
        if key == 'model_matrix':
            for d in self:
                d.model_matrix = value
        else:
            super(Group, self).__setattr__(key, value)

    @staticmethod
    def __assert_drawable(arg):
        assert isinstance(arg, Drawable)

        return arg


# DRAWABLE OBJECTS


class Line(Drawable):
    """
    A path (polyline) made up of line segments.

    Attributes:
        vertices: `array_like`. An array with (x, y, z) coordinates of segment endpoints.
        color: `int`. Packed RGB color of the lines (0xff0000 is red, 0xff is blue).
        width: `float`. The thickness of the lines.
        model_matrix: `array_like`. 4x4 model transform matrix.
    """

    type = Unicode(default_value='Line', read_only=True).tag(sync=True)
    vertices = Array().tag(sync=True, **array_serialization)
    color = Int().tag(sync=True)
    width = Float().tag(sync=True)
    model_matrix = Array().tag(sync=True, **array_serialization)


class MarchingCubes(Drawable):
    """
    An isosurface in a scalar field obtained through Marching Cubes algorithm.

    The default domain of the scalar field is -0.5 < x, y, z < 0.5.
    If the domain should be different, the bounding box needs to be transformed using the model_matrix.

    Attributes:
        scalar_field: `array_like`. A 3D scalar field of values.
        level: `float`. Value at the computed isosurface.
        color: `int`. Packed RGB color of the isosurface (0xff0000 is red, 0xff is blue).
        model_matrix: `array_like`. 4x4 model transform matrix.
    """

    type = Unicode(default_value='MarchingCubes', read_only=True).tag(sync=True)
    scalar_field = Array().tag(sync=True, **array_serialization)
    level = Float().tag(sync=True)
    color = Int().tag(sync=True)
    model_matrix = Array().tag(sync=True, **array_serialization)


class Mesh(Drawable):
    """
    A 3D triangles mesh.

    Attributes:
        vertices: `array_like`. Array of triangle vertices: float (x, y, z) coordinate triplets.
        indices: `array_like`.  Array of vertex indices: int triplets of indices from vertices array.
        color: `int`. Packed RGB color of the mesh (0xff0000 is red, 0xff is blue) when not using color maps.
        attribute: `array_like`. Array of float attribute for the color mapping, coresponding to each vertex.
        color_map: `list`. A list of float quadruplets (attribute value, R, G, B), sorted by attribute value. The first
            quadruplet should have value 0.0, the last 1.0; R, G, B are RGB color components in the range 0.0 to 1.0.
        color_range: `list`. A pair [min_value, max_value], which determines the levels of color attribute mapped
            to 0 and 1 in the color map respectively.
        model_matrix: `array_like`. 4x4 model transform matrix.
    """

    type = Unicode(default_value='Mesh', read_only=True).tag(sync=True)
    vertices = Array().tag(sync=True, **array_serialization)
    indices = Array().tag(sync=True, **array_serialization)
    color = Int().tag(sync=True)
    attribute = Array().tag(sync=True, **array_serialization)
    color_map = Array().tag(sync=True, **array_serialization)
    color_range = ListOrArray(minlen=2, maxlen=2, empty_ok=True).tag(sync=True)
    model_matrix = Array().tag(sync=True, **array_serialization)


class Points(Drawable):
    """
    A point cloud.

    Attributes:
        positions: `array_like`. Array with (x, y, z) coordinates of the points.
        colors: `array_like`. Same-length array of (`int`) packed RGB color of the points (0xff0000 is red, 0xff is blue).
        color: `int`. Packed RGB color of the points (0xff0000 is red, 0xff is blue) when `colors` is empty.
        point_size: `float`. Diameter of the balls representing the points in 3D space.
        shader: `str`. Display style (name of the shader used) of the points.
            Legal values are:
            `flat`: simple circles with uniform color,
            `3d`: little 3D balls,
            `3dSpecular`: little 3D balls with specular lightning,
            `mesh`: high precision triangle mesh of a ball (high quality and GPU load).
        model_matrix: `array_like`. 4x4 model transform matrix.
    """

    type = Unicode(default_value='Points', read_only=True).tag(sync=True)
    positions = Array().tag(sync=True, **array_serialization)
    colors = Array().tag(sync=True, **array_serialization)
    color = Int().tag(sync=True)
    point_size = Float().tag(sync=True)
    shader = Unicode().tag(sync=True)
    model_matrix = Array().tag(sync=True, **array_serialization)

    @validate('colors')
    def _validate_colors(self, proposal):
        required = self.positions.size // 3  # (x, y, z) triplet per 1 color
        actual = proposal['value'].size
        if actual != 0 and required != actual:
            raise TraitError('colors has wrong size: %s (%s required)' % (actual, required))
        return proposal['value']


class STL(Drawable):
    """
    A STereoLitograpy 3D geometry.

    STL is a popular format introduced for 3D printing. There are two sub-formats - ASCII and binary.

    Attributes:
        text: `str`. STL data in text format (ASCII STL).
        binary: `bytes`. STL data in binary format (Binary STL).
            The `text` attribute should be set to None when using Binary STL.
        color: `int`. Packed RGB color of the resulting mesh (0xff0000 is red, 0xff is blue).
        model_matrix: `array_like`. 4x4 model transform matrix.
    """

    type = Unicode(default_value='STL', read_only=True).tag(sync=True)
    text = Unicode(allow_none=True).tag(sync=True)
    binary = Array().tag(sync=True, **array_serialization)
    color = Int().tag(sync=True)
    model_matrix = Array().tag(sync=True, **array_serialization)


class Surface(Drawable):
    """
    Surface plot of a 2D function z = f(x, y).

    The default domain of the scalar field is -0.5 < x, y < 0.5.
    If the domain should be different, the bounding box needs to be transformed using the model_matrix.

    Attributes:
        heights: `array_like`. 2D scalar field of Z values.
        color: `int`. Packed RGB color of the resulting mesh (0xff0000 is red, 0xff is blue).
        model_matrix: `array_like`. 4x4 model transform matrix.
    """

    type = Unicode(default_value='Surface', read_only=True).tag(sync=True)
    heights = Array().tag(sync=True, **array_serialization)
    color = Int().tag(sync=True)
    model_matrix = Array().tag(sync=True, **array_serialization)


class Text(Drawable):
    """
    Text rendered using KaTeX with a 3D position.

    Attributes:
        text: `str`. Content of the text.
        position: `list`. Coordinates (x, y, z) of the text's position.
        color: `int`. Packed RGB color of the text (0xff0000 is red, 0xff is blue).
        reference_point: `str`. Two-letter string representing the text's alignment.
            First letter: 'l', 'c' or 'r': left, center or right
            Second letter: 't', 'c' or 'b': top, center or bottom.
        size: `float`. Font size in 'em' HTML units.
    """

    type = Unicode(default_value='Text', read_only=True).tag(sync=True)
    text = Unicode().tag(sync=True)
    position = ListOrArray(minlen=3, maxlen=3).tag(sync=True)
    color = Int().tag(sync=True)
    reference_point = Unicode().tag(sync=True)
    size = Float().tag(sync=True)


class Text2d(Drawable):
    """
    Text rendered using KaTeX with a fixed 2D position, independent of camera settings.

    Attributes:
        text: `str`. Content of the text.
        position: `list`. Ratios (r_x, r_y) of the text's position in range (0, 1) - relative to canvas size.
        color: `int`. Packed RGB color of the text (0xff0000 is red, 0xff is blue).
        reference_point: `str`. Two-letter string representing the text's alignment.
            First letter: 'l', 'c' or 'r': left, center or right
            Second letter: 't', 'c' or 'b': top, center or bottom.
        size: `float`. Font size in 'em' HTML units.
    """

    type = Unicode(default_value='Text2d', read_only=True).tag(sync=True)
    color = Int().tag(sync=True)
    size = Float().tag(sync=True)
    reference_point = Unicode().tag(sync=True)
    position = ListOrArray(minlen=2, maxlen=2).tag(sync=True)
    text = Unicode().tag(sync=True)


class Texture(Drawable):
    """
    A 2D image displayed as a texture.

    By default, the texture image is mapped into the square: -0.5 < x, y < 0.5, z = 1.
    If the size (scale, aspect ratio) or position should be different then the texture should be transformed
    using the model_matrix.

    Attributes:
        binary: `bytes`. Image data in a specific format.
        file_format: `str`. Format of the data, it should be the second part of MIME format of type 'image/',
            for example 'jpeg', 'png', 'gif', 'tiff'.
        model_matrix: `array_like`. 4x4 model transform matrix.
    """

    type = Unicode(default_value='Texture', read_only=True).tag(sync=True)
    binary = Bytes().tag(sync=True)
    file_format = Unicode().tag(sync=True)
    model_matrix = Array().tag(sync=True, **array_serialization)


class TextureText(Drawable):
    """
    A text in the 3D space rendered using a texture.

    Compared to Text and Text2d this drawable has less features (no KaTeX support), but the labels are located
    in the GPU memory, and not the browser's DOM tree. This has performance consequences, and may be preferable when
    many simple labels need to be displayed.

    Attributes:
        text: `str`. Content of the text.
        position: `list`. Coordinates (x, y, z) of the text's position.
        color: `int`. Packed RGB color of the text (0xff0000 is red, 0xff is blue).
        size: `float`. Size of the texture sprite containing the text.
        font_face: `str`. Name of the font to use for rendering the text.
        font_weight: `int`. Thickness of the characters in HTML-like units from the range (100, 900), where
            400 is normal and 600 is bold font.
        font_size: `int`. The font size inside the sprite texture in px units. This does not affect the size of the
            text in the scene, only the accuracy and raster size of the texture.
    """

    type = Unicode(default_value='TextureText', read_only=True).tag(sync=True)
    text = Unicode().tag(sync=True)
    position = ListOrArray(minlen=3, maxlen=3).tag(sync=True)
    color = Int().tag(sync=True)
    size = Float().tag(sync=True)
    font_face = Unicode().tag(sync=True)
    font_weight = Int().tag(sync=True)
    font_size = Int().tag(sync=True)


class VectorField(Drawable):
    """
    A dense 3D or 2D vector field.

    By default, the origins of the vectors are assumed to be a grid inscribed in the -0.5 < x, y, z < 0.5 cube
    or -0.5 < x, y < 0.5 square, regardless of the passed vector field shape (aspect ratio etc.).
    Different grid size, shape and rotation can be obtained using the model_matrix.

    The color of the vectors is a gradient from origin_color to head_color. Heads, when used, have uniform head_color.

    For sparse (i.e. not forming a grid) 3D vectors, use the `Vectors` drawable.

    Attributes:
        vectors: `array_like`. Vector field of shape (L, H, W, 3) for 3D fields or (H, W, 2) for 2D fields.
        colors: `array_like`. Twice the length of vectors array of int: packed RGB colors
            (0xff0000 is red, 0xff is blue).
            The array has consecutive pairs (origin_color, head_color) for vectors in row-major order.
        origin_color: `int`. Packed RGB color of the origins (0xff0000 is red, 0xff is blue) when `colors` is empty.
        head_color: `int`. Packed RGB color of the vector heads (0xff0000 is red, 0xff is blue) when `colors` is empty.
        use_head: `bool`. Whether vectors should display an arrow head.
        head_size: `float`. The size of the arrow heads.
        scale: `float`. Scale factor for the vector lengths, for artificially scaling the vectors in place.
        model_matrix: `array_like`. 4x4 model transform matrix.
    """

    type = Unicode(default_value='VectorField', read_only=True).tag(sync=True)
    vectors = Array().tag(sync=True, **array_serialization)
    colors = Array().tag(sync=True, **array_serialization)
    origin_color = Int().tag(sync=True)
    head_color = Int().tag(sync=True)
    use_head = Bool().tag(sync=True)
    head_size = Float().tag(sync=True)
    scale = Float().tag(sync=True)
    model_matrix = Array().tag(sync=True, **array_serialization)

    @validate('vectors')
    def _validate_vectors(self, proposal):
        shape = proposal['value'].shape
        if len(shape) not in (3, 4) or len(shape) != shape[-1] + 1:
            raise TraitError('Vector field has invalid shape: {}, '
                             'expected (L, H, W, 3) for a 3D or (H, W, 2) for a 2D field'.format(shape))
        return np.array(proposal['value'], np.float32)


class Vectors(Drawable):
    """
    3D vectors.

    The color of the vectors is a gradient from origin_color to head_color. Heads, when used, have uniform head_color.

    For dense (i.e. forming a grid) 3D or 2D vectors, use the `VectorField` drawable.

    Attributes:
        vectors: `array_like`. The vectors as (dx, dy, dz) float triples.
        origins: `array_like`. Same-size array of (x, y, z) coordinates of vector origins.
        colors: `array_like`. Twice the length of vectors array of int: packed RGB colors
            (0xff0000 is red, 0xff is blue).
            The array has consecutive pairs (origin_color, head_color) for vectors in row-major order.
        origin_color: `int`. Packed RGB color of the origins (0xff0000 is red, 0xff is blue), default: same as color.
        head_color: `int`. Packed RGB color of the vector heads (0xff0000 is red, 0xff is blue), default: same as color.
        use_head: `bool`. Whether vectors should display an arrow head.
        head_size: `float`. The size of the arrow heads.
        labels: `list` of `str`. Captions to display next to the vectors.
        label_size: `float`. Label font size in 'em' HTML units.
        line_width: `float`. Width of the vector segments.
        model_matrix: `array_like`. 4x4 model transform matrix.
    """

    type = Unicode(default_value='Vectors', read_only=True).tag(sync=True)
    origins = Array().tag(sync=True, **array_serialization)
    vectors = Array().tag(sync=True, **array_serialization)
    colors = Array().tag(sync=True, **array_serialization)
    origin_color = Int().tag(sync=True)
    head_color = Int().tag(sync=True)
    use_head = Bool().tag(sync=True)
    head_size = Float().tag(sync=True)
    labels = List().tag(sync=True)
    label_size = Float().tag(sync=True)
    line_width = Float().tag(sync=True)
    model_matrix = Array().tag(sync=True, **array_serialization)


class Voxels(Drawable):
    """
    3D volumetric data.

    By default, the voxels are a grid inscribed in the -0.5 < x, y, z < 0.5 cube
    regardless of the passed voxel array shape (aspect ratio etc.).
    Different grid size, shape and rotation can be obtained using the model_matrix.

    Attributes:
        voxels: `array_like`. 3D array of `int` in range (0, 255).
            0 means empty voxel, 1 and above refer to consecutive color_map entries.
        color_map: `array_like`. Flat array of `int` packed RGB colors (0xff0000 is red, 0xff is blue).
            The color defined at index i is for voxel value (i+1), e.g.:
            color_map = [0xff, 0x00ff]
            voxels = [[[
                0, # empty voxel
                1, # blue voxel
                2  # red voxel
            ]]]
        model_matrix: `array_like`. 4x4 model transform matrix.
    """

    type = Unicode(default_value='Voxels', read_only=True).tag(sync=True)
    voxels = Array().tag(sync=True, **array_serialization)
    color_map = Array().tag(sync=True, **array_serialization)
    model_matrix = Array().tag(sync=True, **array_serialization)



import unittest
import numpy as np
import k3d
from ..transform import Transform


class TestTransform(unittest.TestCase):
    def test_recompute_translation(self):
        # given
        transform = Transform()
        # when
        transform.translation = [1., 2., 3.]
        # then
        self.assertTrue((transform.model_matrix == np.array([
            [1., 0., 0., 1.],
            [0., 1., 0., 2.],
            [0., 0., 1., 3.],
            [0., 0., 0., 1.]
        ])).all())

    def test_recompute_scaling(self):
        # given
        transform = Transform()
        # when
        transform.scaling = [1., 2., 3.]
        # then
        self.assertTrue((transform.model_matrix == np.array([
            [1., 0., 0., 0.],
            [0., 2., 0., 0.],
            [0., 0., 3., 0.],
            [0., 0., 0., 1.]
        ])).all())

    def test_drawable_notification(self):
        # given
        transform = Transform()
        points = k3d.points([0, 0, 1])
        transform.add_drawable(points)
        # when
        transform.scaling = [1., 2., 3.]
        # then
        self.assertTrue((points.model_matrix == np.array([
            [1., 0., 0., 0.],
            [0., 2., 0., 0.],
            [0., 0., 3., 0.],
            [0., 0., 0., 1.]
        ])).all())


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np

from ..helpers import map_colors
from ..colormaps.basic_color_maps import Rainbow, Binary


class TestPythonColorMapping(unittest.TestCase):
    # Rainbow: for 0 - blue, for 1 - red

    def test_zeros(self):
        # given
        attribute = np.zeros(5)
        # when
        colors = map_colors(attribute, Rainbow)
        # then
        self.assertTrue((colors == np.ones(5, dtype=np.int32) * 0xff).all())

    def test_ones(self):
        # given
        attribute = np.ones(5)
        # when
        # color range needed here, if uniform, range inferred as [1., 2.], which also gives all blue
        colors = map_colors(attribute, Rainbow, color_range=(0., 1.))
        # then
        self.assertTrue((colors == np.ones(5, dtype=np.int32) * 0xff0000).all())

    def test_gradient(self):
        # given
        attribute = np.array([0, 0.5, 1])
        # when
        colors = map_colors(attribute, Binary)
        # then
        self.assertTrue((colors == [0xffffff, 0x7f7f7f, 0]).all())


if __name__ == '__main__':
    unittest.main()



from ._version import version_info

from .colormaps import paraview_color_maps
from .colormaps import basic_color_maps
from .colormaps import matplotlib_color_maps

from .k3d import (plot,
                  line,
                  marching_cubes,
                  mesh,
                  points,
                  stl,
                  surface,
                  text,
                  text2d,
                  texture,
                  texture_text,
                  vector_field,
                  vectors,
                  voxels,
                  vtk_poly_data)


def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'static',
        'dest': 'k3d',
        'require': 'k3d/extension'
    }]



# optional dependency
try:
    # noinspection PyPackageRequirements
    import vtk
    # noinspection PyPackageRequirements
    from vtk.util import numpy_support
except ImportError:
    vtk = None
    numpy_support = None

import numpy as np
import six
from .colormaps import basic_color_maps
from .plot import Plot
from .objects import (Line, MarchingCubes, Mesh, Points, STL, Surface, Text, Text2d, Texture, TextureText, VectorField,
                      Vectors, Voxels)
from .transform import process_transform_arguments
from .helpers import check_attribute_range

_default_color = 0x0000FF  # blue


def line(vertices, color=_default_color, width=1, **kwargs):
    """Create a Line drawable for plotting segments and polylines.

    Arguments:
        vertices: `array_like`. Array with (x, y, z) coordinates of segment endpoints.
        color: `int`. Packed RGB color of the lines (0xff0000 is red, 0xff is blue).
        width: `float`. Thickness of the lines.
        kwargs: `dict`. Dictionary arguments to configure transform and model_matrix."""
    return process_transform_arguments(
        Line(vertices=np.array(vertices, np.float32), color=color, width=width),
        **kwargs
    )


def marching_cubes(scalar_field, level, color=_default_color, **kwargs):
    """Create a MarchingCubes drawable.

    Plot an isosurface of a scalar field obtained through a Marching Cubes algorithm.

    The default domain of the scalar field is -0.5 < x, y, z < 0.5.

    If the domain should be different, the bounding box needs to be transformed using kwargs, like this:
        marching_cubes(..., bounds=[-1, 1, -1, 1, -1, 1])
    or:
        marching_cubes(..., xmin=-10, xmax=10, ymin=-4, ymax=4, zmin=0, zmax=20)
    or:
        marching_cubes(..., scaling=[width, height, length])

    Arguments:
        scalar_field: `array_like`. A 3D scalar field of values.
        level: `float`. Value at the computed isosurface.
        color: `int`. Packed RGB color of the isosurface (0xff0000 is red, 0xff is blue).
        kwargs: `dict`. Dictionary arguments to configure transform and model_matrix."""
    return process_transform_arguments(
        MarchingCubes(scalar_field=np.array(scalar_field, np.float32),
                      color=color,
                      level=level),
        **kwargs
    )


def mesh(vertices, indices, color=_default_color, attribute=(), color_map=(), color_range=(), **kwargs):
    """Create a Mesh drawable representing a 3D triangles mesh.

    Arguments:
        vertices: `array_like`. Array of triangle vertices: float (x, y, z) coordinate triplets.
        indices: `array_like`.  Array of vertex indices: int triplets of indices from vertices array.
        color: `int`. Packed RGB color of the mesh (0xff0000 is red, 0xff is blue) when not using color maps.
        attribute: `array_like`. Array of float attribute for the color mapping, coresponding to each vertex.
        color_map: `list`. A list of float quadruplets (attribute value, R, G, B), sorted by attribute value. The first
            quadruplet should have value 0.0, the last 1.0; R, G, B are RGB color components in the range 0.0 to 1.0.
        color_range: `list`. A pair [min_value, max_value], which determines the levels of color attribute mapped
            to 0 and 1 in the color map respectively.
        kwargs: `dict`. Dictionary arguments to configure transform and model_matrix."""
    color_map = np.array(color_map, np.float32)
    attribute = np.array(attribute, np.float32)
    color_range = check_attribute_range(attribute, color_range)

    return process_transform_arguments(
        Mesh(vertices=np.array(vertices, np.float32),
             indices=np.array(indices, np.uint32),
             color=color,
             attribute=attribute,
             color_map=color_map,
             color_range=color_range),
        **kwargs
    )


def points(positions, colors=(), color=_default_color, point_size=1.0, shader='3dSpecular', **kwargs):
    """Create a Points drawable representing a point cloud.

    Arguments:
        positions: `array_like`. Array with (x, y, z) coordinates of the points.
        colors: `array_like`. Same-length array of `int`-packed RGB color of the points (0xff0000 is red, 0xff is blue).
        color: `int`. Packed RGB color of the points (0xff0000 is red, 0xff is blue) when `colors` is empty.
        point_size: `float`. Diameter of the balls representing the points in 3D space.
        shader: `str`. Display style (name of the shader used) of the points.
            Legal values are:
            `flat`: simple circles with uniform color,
            `3d`: little 3D balls,
            `3dSpecular`: little 3D balls with specular lightning,
            `mesh`: high precision triangle mesh of a ball (high quality and GPU load).
        kwargs: `dict`. Dictionary arguments to configure transform and model_matrix."""
    return process_transform_arguments(
        Points(positions=np.array(positions, np.float32), colors=np.array(colors, np.float32),
               color=color, point_size=point_size, shader=shader),
        **kwargs
    )


# noinspection PyShadowingNames
def stl(stl, color=_default_color, **kwargs):
    """Create an STL drawable for data in STereoLitograpy format.

    Arguments:
        stl: `str` or `bytes`. STL data in either ASCII STL (string) or Binary STL (bytes).
        color: `int`. Packed RGB color of the resulting mesh (0xff0000 is red, 0xff is blue).
        kwargs: `dict`. Dictionary arguments to configure transform and model_matrix."""
    plain = isinstance(stl, six.string_types)
    return process_transform_arguments(
        STL(text=stl if plain else None,
            binary=([]  # allow_null doesn't really work for Array...
                    if plain
                    else np.frombuffer(stl, dtype=np.uint8)),
            color=color),
        **kwargs
    )


def surface(heights, color=_default_color, **kwargs):
    """Create a Surface drawable.

    Plot a 2d function: z = f(x, y).

    The default domain of the scalar field is -0.5 < x, y < 0.5.

    If the domain should be different, the bounding box needs to be transformed using kwargs, like this:
        surface(..., bounds=[-1, 1, -1, 1])
    or:
        surface(..., xmin=-10, xmax=10, ymin=-4, ymax=4)

    Arguments:
        heights: `array_like`. A 2d scalar function values grid.
        color: `int`. Packed RGB color of the surface (0xff0000 is red, 0xff is blue).
        kwargs: `dict`. Dictionary arguments to configure transform and model_matrix."""
    return process_transform_arguments(
        Surface(heights=np.array(heights, np.float32), color=color),
        **kwargs
    )


# noinspection PyShadowingNames
def text(text, position=(0, 0, 0), color=_default_color, reference_point='lb', size=1.0):
    """Create a Text drawable for 3D-positioned text labels.

    Arguments:
        text: `str`. Content of the text.
        position: `list`. Coordinates (x, y, z) of the text's position.
        color: `int`. Packed RGB color of the text (0xff0000 is red, 0xff is blue).
        reference_point: `str`. Two-letter string representing the text's alignment.
            First letter: 'l', 'c' or 'r': left, center or right
            Second letter: 't', 'c' or 'b': top, center or bottom.
        size: `float`. Font size in 'em' HTML units."""
    return Text(position=position, reference_point=reference_point, text=text, size=size, color=color)


# noinspection PyShadowingNames
def text2d(text, position=(0, 0), color=_default_color, size=1.0, reference_point='lt'):
    """Create a Text2d drawable for 2D-positioned (viewport bound, OSD) labels.

    Arguments:
        text: `str`. Content of the text.
        position: `list`. Ratios (r_x, r_y) of the text's position in range (0, 1) - relative to canvas size.
        color: `int`. Packed RGB color of the text (0xff0000 is red, 0xff is blue).
        reference_point: `str`. Two-letter string representing the text's alignment.
            First letter: 'l', 'c' or 'r': left, center or right
            Second letter: 't', 'c' or 'b': top, center or bottom.
        size: `float`. Font size in 'em' HTML units."""
    return Text2d(position=position, reference_point=reference_point, text=text, size=size, color=color)


def texture(binary, file_format, **kwargs):
    """Create a Texture drawable for displaying 2D raster images in common formats.

    By default, the texture image is mapped into the square: -0.5 < x, y < 0.5, z = 1.
    If the size (scale, aspect ratio) or position should be different then the texture should be transformed
    using kwargs, for example:
        texture(..., xmin=0, xmax=640, ymin=0, ymax=480)
    or:
        texture(..., bounds=[0, 10, 0, 20])
    or:
        texture(..., scaling=[1.0, 0.75, 0])

    Arguments:
        binary: `bytes`. Image data in a specific format.
        file_format: `str`. Format of the data, it should be the second part of MIME format of type 'image/',
            for example 'jpeg', 'png', 'gif', 'tiff'.
        kwargs: `dict`. Dictionary arguments to configure transform and model_matrix."""
    return process_transform_arguments(
        Texture(binary=binary,
                file_format=file_format),
        **kwargs
    )


# noinspection PyShadowingNames
def texture_text(text, position=(0, 0, 0), color=_default_color, font_weight=400, font_face='Courier New',
                 font_size=68, size=1.0):
    """Create a TextureText drawable.

    Compared to Text and Text2d this drawable has less features (no KaTeX support), but the labels are located
    in the GPU memory, and not the browser's DOM tree. This has performance consequences, and may be preferable when
    many simple labels need to be displayed.

    Arguments:
        text: `str`. Content of the text.
        position: `list`. Coordinates (x, y, z) of the text's position.
        color: `int`. Packed RGB color of the text (0xff0000 is red, 0xff is blue).
        size: `float`. Size of the texture sprite containing the text.
        font_face: `str`. Name of the font to use for rendering the text.
        font_weight: `int`. Thickness of the characters in HTML-like units from the range (100, 900), where
            400 is normal and 600 is bold font.
        font_size: `int`. The font size inside the sprite texture in px units. This does not affect the size of the
            text in the scene, only the accuracy and raster size of the texture."""
    return TextureText(text=text, position=position, color=color, size=size,
                       font_face=font_face, font_size=font_size, font_weight=font_weight)


# noinspection PyShadowingNames
def vector_field(vectors,
                 colors=(),
                 origin_color=None, head_color=None, color=_default_color,
                 use_head=True, head_size=1.0, scale=1.0, **kwargs):
    """Create a VectorField drawable for displaying dense 2D or 3D grids of vectors of same dimensionality.

    By default, the origins of the vectors are assumed to be a grid inscribed in the -0.5 < x, y, z < 0.5 cube
    or -0.5 < x, y < 0.5 square, regardless of the passed vector field shape (aspect ratio etc.).
    Different grid size, shape and rotation can be obtained using kwargs:
        vector_field(..., bounds=[-pi, pi, -pi, pi, 0, 1])
    or:
        vector_field(..., scaling=[scale_x, scale_y, scale_z]).

    For sparse (i.e. not forming a grid) 3D vectors, use the `vectors()` function.

    Arguments:
        vectors: `array_like`. Vector field of shape (L, H, W, 3) for 3D fields or (H, W, 2) for 2D fields.
        colors: `array_like`. Twice the length of vectors array of int: packed RGB colors
            (0xff0000 is red, 0xff is blue).
            The array has consecutive pairs (origin_color, head_color) for vectors in row-major order.
        origin_color: `int`. Packed RGB color of the origins (0xff0000 is red, 0xff is blue), default: same as color.
        head_color: `int`. Packed RGB color of the vector heads (0xff0000 is red, 0xff is blue), default: same as color.
        color: `int`. Packed RGB color of the vectors (0xff0000 is red, 0xff is blue) when `colors` is empty and
            origin_color and head_color are not specified.
        use_head: `bool`. Whether vectors should display an arrow head.
        head_size: `float`. The size of the arrow heads.
        scale: `float`. Scale factor for the vector lengths, for artificially scaling the vectors in place.
        kwargs: `dict`. Dictionary arguments to configure transform and model_matrix."""
    return process_transform_arguments(
        VectorField(vectors=vectors,
                    colors=np.array(colors, np.uint32),
                    use_head=use_head,
                    head_size=head_size,
                    head_color=head_color if head_color is not None else color,
                    origin_color=origin_color if origin_color is not None else color,
                    scale=scale),
        **kwargs
    )


# noinspection PyShadowingNames
def vectors(origins, vectors=None, colors=(),
            origin_color=None, head_color=None, color=_default_color,
            use_head=True, head_size=1.0,
            labels=(), label_size=1.0,
            line_width=1, **kwargs):
    """Create a Vectors drawable representing individual 3D vectors.

    The color of the vectors is a gradient from origin_color to head_color. Heads, when used, have uniform head_color.

    For dense (i.e. forming a grid) 3D or 2D vectors, use the `vector_field` function.

    Arguments:
        origins: `array_like`. Array of (x, y, z) coordinates of vector origins, when `vectors` is None, these
            are (dx, dy, dz) components of unbound vectors (which are displayed as originating in (0, 0, 0)).
        vectors: `array_like`. The vectors as (dx, dy, dz) float triples. When not given, the `origins` are taken
            as vectors. When given, it must be same size as `origins`.
        colors: `array_like`. Twice the length of vectors array of int: packed RGB colors
            (0xff0000 is red, 0xff is blue).
            The array has consecutive pairs (origin_color, head_color) for vectors in row-major order.
        origin_color: `int`. Packed RGB color of the origins (0xff0000 is red, 0xff is blue), default: same as color.
        head_color: `int`. Packed RGB color of the vector heads (0xff0000 is red, 0xff is blue), default: same as color.
        color: `int`. Packed RGB color of the vectors (0xff0000 is red, 0xff is blue) when `colors` is empty and
            origin_color and head_color are not specified.
        use_head: `bool`. Whether vectors should display an arrow head.
        head_size: `float`. The size of the arrow heads.
        labels: `list` of `str`. Captions to display next to the vectors.
        label_size: `float`. Label font size in 'em' HTML units.
        line_width: `float`. Width of the vector segments.
        kwargs: `dict`. Dictionary arguments to configure transform and model_matrix."""
    return process_transform_arguments(
        Vectors(
            vectors=vectors if vectors is not None else origins,
            origins=origins if vectors is not None else np.zeros_like(vectors),
            colors=np.array(colors, np.uint32),
            origin_color=origin_color if origin_color is not None else color,
            head_color=head_color if head_color is not None else color,
            use_head=use_head,
            head_size=head_size,
            labels=labels,
            label_size=label_size,
            line_width=line_width
        ),
        **kwargs
    )


# noinspection PyShadowingNames
def voxels(voxels, color_map, **kwargs):
    """Create a Voxels drawable for 3D volumetric data.

    By default, the voxels are a grid inscribed in the -0.5 < x, y, z < 0.5 cube
    regardless of the passed voxel array shape (aspect ratio etc.).
    Different grid size, shape and rotation can be obtained using  kwargs:
        voxels(..., bounds=[0, 300, 0, 400, 0, 500])
    or:
        voxels(..., scaling=[scale_x, scale_y, scale_z]).

    Arguments:
        voxels: `array_like`. 3D array of `int` in range (0, 255).
            0 means empty voxel, 1 and above refer to consecutive color_map entries.
        color_map: `array_like`. Flat array of `int` packed RGB colors (0xff0000 is red, 0xff is blue).
            The color defined at index i is for voxel value (i+1), e.g.:
            color_map = [0xff, 0x00ff]
            voxels = [[[
                0, # empty voxel
                1, # blue voxel
                2  # red voxel
            ]]]
    kwargs: `dict`. Dictionary arguments to configure transform and model_matrix."""
    return process_transform_arguments(
        Voxels(voxels=np.array(voxels, np.uint8), color_map=np.array(color_map, np.float32)),
        **kwargs
    )


def vtk_poly_data(poly_data, color=_default_color, color_attribute=None, color_map=basic_color_maps.Rainbow, **kwargs):
    """Create a Mesh drawable from given vtkPolyData.

    This function requires the vtk module (from package VTK) to be installed.

    Arguments:
        poly_data: `vtkPolyData`. Native vtkPolyData geometry.
        color: `int`. Packed RGB color of the resulting mesh (0xff0000 is red, 0xff is blue) when not using color maps.
        color_attribute: `tuple` of (`str`, `float`, `float`). This determines which scalar should be taken as the
            attribute for the color_map, and the color_range for the mesh: (attribute_name, min_value, max_value).
            A VTK mesh can have multiple named attributes in the vertices.
            min_value is the value mapped to 0 in the color_map.
            max_value is the value mapped to 1 in the color_map.
        color_map: `list`. A list of float quadruplets (attribute value, R, G, B), sorted by attribute value. The first
            quadruplet should have value 0.0, the last 1.0; R, G, B are RGB color components in the range 0.0 to 1.0.
        kwargs: `dict`. Dictionary arguments to configure transform and model_matrix."""
    if vtk is None:
        raise RuntimeError('vtk module is not available')

    if poly_data.GetPolys().GetMaxCellSize() > 3:
        cut_triangles = vtk.vtkTriangleFilter()
        cut_triangles.SetInputData(poly_data)
        cut_triangles.Update()
        poly_data = cut_triangles.GetOutput()

    if color_attribute is not None:
        attribute = numpy_support.vtk_to_numpy(poly_data.GetPointData().GetArray(color_attribute[0]))
        color_range = color_attribute[1:3]
    else:
        attribute = []
        color_range = []

    vertices = numpy_support.vtk_to_numpy(poly_data.GetPoints().GetData())
    indices = numpy_support.vtk_to_numpy(poly_data.GetPolys().GetData()).reshape(-1, 4)[:, 1:4]

    return process_transform_arguments(
        Mesh(vertices=np.array(vertices, np.float32),
             indices=np.array(indices, np.uint32),
             color=color,
             attribute=np.array(attribute, np.float32),
             color_range=color_range,
             color_map=np.array(color_map, np.float32)),
        **kwargs
    )


def plot(height=512,
         antialias=True,
         background_color=0xffffff,
         camera_auto_fit=True,
         grid_auto_fit=True,
         voxel_paint_color=0,
         grid=(-1, -1, -1, 1, 1, 1)):
    """Create a K3D Plot widget.

    This creates the main widget for displaying 3D objects.

    Arguments:
        height: `int`. Height of the widget in pixels.
        antialias: `bool`. Enable antialiasing in WebGL renderer.
        background_color: `int`.  Packed RGB color of the plot background (0xff0000 is red, 0xff is blue).
        camera_auto_fit: `bool`. Enable automatic camera setting after adding, removing or changing a plot object.
        grid_auto_fit: `bool`. Enable automatic adjustment of the plot grid to contained objects.
        voxel_paint_color: `int`. The (initial) int value to be inserted when editing voxels.
        grid: `array_like`. 6-element tuple specifying the bounds of the plot grid (x0, y0, z0, x1, y1, z1)."""
    return Plot(antialias=antialias,
                background_color=background_color,
                camera_auto_fit=camera_auto_fit, grid_auto_fit=grid_auto_fit,
                height=height,
                voxel_paint_color=voxel_paint_color, grid=grid)



import urllib.request
import xml.etree.ElementTree as ET

response = urllib.request.urlopen('http://www.paraview.org/Wiki/images/d/d4/All_mpl_cmaps.xml')
root = ET.fromstring(response.read().decode('utf8'))

file = open('matplotlib_color_maps.py', 'w')

for colorMap in root:
    name = colorMap.attrib['name']
    name = name[:1].upper() + name[1:]
    file.write(name + ' = [\n')

    for p in colorMap:
        file.write('    ' + p.attrib['x'] + ', ' + p.attrib['r'] + ', ' + p.attrib['g'] + ', '
                   + p.attrib['b'] + ',\n')

    file.write(']\n\n')

file.close()



from __future__ import print_function

import os

from setuptools import setup, find_packages

from distutils import log

from setupbase import (
    create_cmdclass, install_npm, combine_commands, BaseCommand,
    ensure_targets
)

here = os.path.dirname(os.path.abspath(__file__))
node_root = os.path.join(here, 'js')

log.set_verbosity(log.DEBUG)
log.info('setup.py entered')
log.info('$PATH=%s' % os.environ['PATH'])

LONG_DESCRIPTION = 'Jupyter notebook extension for 3D visualization.'

# Representative files that should exist after a successful build
targets = [
    os.path.join(here, 'k3d', 'static', 'extension.js'),
    os.path.join(here, 'k3d', 'static', 'index.js')
]

cmdclass = create_cmdclass(('jsdeps',))
cmdclass['jsdeps'] = combine_commands(
    install_npm(node_root),
    ensure_targets(targets)
)

version_ns = {}
with open(os.path.join(here, 'k3d', '_version.py')) as f:
    exec(f.read(), {}, version_ns)

setup_args = {
    'name': 'K3D',
    'version': version_ns['__version__'],
    'license': 'MIT',
    'description': 'Jupyter notebook extension for 3D visualization.',
    'long_description': LONG_DESCRIPTION,
    'include_package_data': True,
    'data_files': [
        ('share/jupyter/nbextensions/k3d', [
            'k3d/static/*'
        ]),
    ],
    'install_requires': [
        'ipywidgets>=7.0.1',
        'traittypes',
        'traitlets',
        'numpy>=1.11.0'
    ],
    'packages': find_packages(),
    'zip_safe': False,
    'cmdclass': cmdclass,
    'author': 'k3d team',
    'author_email': 'artur.trzesiok@gmail.com',
    'url': 'http://jupyter.org',
    'keywords': [
        'ipython',
        'jupyter',
        'widgets',
    ],
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Framework :: IPython',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Multimedia :: Graphics',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
}

setup(**setup_args)



import numpy as np
from functools import reduce
import weakref

_epsilon = 1e-6


def get_bounds_fit_matrix(xmin, xmax, ymin, ymax, zmin, zmax):
    """Create a 4x4 transform matrix which maps the default bounding box ([-0.5, 0.5] in all dimensions) into
    a custom bounding box ([xmin, xmax, ymin, ymax, zmin, zmax]).

    It is used for fitting fields (VectorField, Surface, MachingCubes) into custom domains."""

    for name, value in locals().items():
        try:
            float(value)
        except (TypeError, ValueError):
            raise TypeError('%s: expected float, %s given' % (name, type(value).__name__))

    matrix = np.diagflat(np.array((xmax - xmin, ymax - ymin, zmax - zmin, 1.0), np.float32, order='C'))
    matrix[0:3, 3] = ((xmax + xmin) / 2.0, (ymax + ymin) / 2.0, (zmax + zmin) / 2.0)

    return matrix


class Transform(object):
    """
    Abstraction of a 4x4 model transformation matrix with hierarchy support.
    """

    def __init__(self, bounds=None, translation=None, rotation=None, scaling=None, custom_matrix=None, parent=None):
        self.bounds = bounds
        self.translation = translation
        self.rotation = rotation
        self.scaling = scaling
        self.parent = parent
        self.drawables = []
        self.children = []
        self.parent_matrix = parent.model_matrix if parent else np.identity(4)
        self.custom_matrix = custom_matrix if custom_matrix is not None else np.identity(4)
        self.model_matrix = np.identity(4)
        self._recompute_matrix()

    def __setattr__(self, key, value):
        """Set attributes with conversion to ndarray where needed."""
        is_set = hasattr(self, key)  # == False in constructor

        # parameter canonicalization and some validation via reshaping
        if value is None:
            # TODO: maybe forbid for some fields
            pass
        elif key == 'translation':
            value = np.array(value).reshape(3, 1)
        elif key == 'rotation':
            value = np.array(value).reshape(4)
            if not -1 <= value[0] <= 1:
                raise ValueError('Cosine of Theta/2 oustide of [-1; 1] range')
            norm = np.linalg.norm(value[1:4])
            needed_norm = np.sqrt(1 - value[0] * value[0])
            if abs(norm - needed_norm) > _epsilon:
                if norm < _epsilon:
                    raise ValueError('Norm of (x, y, z) part of quaternion too close to zero')
                value[1:4] = value[1:4] / norm * needed_norm
        elif key == 'scaling':
            value = np.array(value).reshape(3)
        elif key in ['parent_matrix', 'custom_matrix', 'model_matrix']:
            value = np.array(value).reshape((4, 4))

        super(Transform, self).__setattr__(key, value)

        if is_set and key != 'model_matrix':
            self._recompute_matrix()
            self._notify_dependants()

    def _recompute_matrix(self):
        # this method shouldn't modify any fields except self.model_matrix

        if self.bounds is None or len(self.bounds) == 0:
            fit_matrix = np.identity(4)
        else:
            if len(self.bounds) == 6:
                xmin, xmax, ymin, ymax, zmin, zmax = self.bounds
            elif len(self.bounds) == 4:
                xmin, xmax, ymin, ymax = self.bounds
                zmin, zmax = -0.5, 0.5
            elif len(self.bounds) == 2:
                # 1-D bounds for a 2D strip - why not?
                xmin, xmax = self.bounds
                ymin, ymax, zmin, zmax = -0.5, 0.5, -0.5, 0.5
            else:
                raise ValueError('Wrong size of bounds array ({}), should be 4 for 2D or 6 for 3D bounds.'.format(
                    self.bounds
                ))
            fit_matrix = get_bounds_fit_matrix(xmin, xmax, ymin, ymax, zmin, zmax)

        if self.translation is not None:
            translation_matrix = np.vstack((
                np.hstack((np.identity(3), np.array(self.translation).reshape(3, 1))),
                np.array([0., 0., 0., 1.]).reshape(1, 4)
            ))
        else:
            translation_matrix = np.identity(4)

        if self.rotation is not None:
            a, b, c, d = self.rotation
            rotation_matrix = np.array([
                [a * a + b * b - c * c - d * d, 2 * (b * c - a * d), 2 * (b * d + a * c), 0.],
                [2 * (b * c + a * d), a * a - b * b + c * c - d * d, 2 * (c * d - a * b), 0.],
                [2 * (b * d - a * c), 2 * (c * d - a * b), a * a - b * b - c * c + d * d, 0.],
                [0., 0., 0., 1.]
            ])
        else:
            rotation_matrix = np.identity(4)

        if self.scaling is not None:
            scaling_matrix = np.diag(np.append(self.scaling, 1.0))
        else:
            scaling_matrix = np.identity(4)

        self.model_matrix = reduce(np.dot, [
            fit_matrix, translation_matrix, rotation_matrix, scaling_matrix, self.custom_matrix, self.parent_matrix
        ])

    def _add_child(self, transform):
        self.children.append(weakref.ref(transform))

    def add_drawable(self, drawable):
        """Register a Drawable to have its model_matrix overwritten after changes to the transform or its parent."""
        self.drawables.append(weakref.ref(drawable))

    def parent_updated(self):
        """Read updated parent transform matrix and update own model_matrix.

        This method should be normally only called by parent Transform to notify its children.
        """
        if self.parent is not None:
            self.parent_matrix = self.parent.model_matrix
        self._recompute_matrix()
        self._notify_dependants()

    def _notify_dependants(self):
        extant_children = []
        for child_ref in self.children:
            child = child_ref()
            if child is not None:
                child.parent_updated()
                extant_children.append(child_ref)
        self.children[:] = extant_children

        extant_drawables = []
        for drawable_ref in self.drawables:
            drawable = drawable_ref()
            if drawable is not None:
                drawable.model_matrix = self.model_matrix
                extant_drawables.append(drawable_ref)
        self.drawables[:] = extant_drawables


def process_transform_arguments(drawable, **kwargs):
    """Process keyword arguments dictionary for a drawable to create a Transform for it.

    Keyword arguments:
        transform: `Transform`. An existing transform object to be (re-)used for the drawable.
            This is useful if the transform should have a parent.
        xmin: `float`. Lower bound in the X dimension for array fields of scalars or vectors.
        xmax: `float`. Upper bound in the X dimension for array fields of scalars or vectors.
        ymin: `float`. Lower bound in the Y dimension for array fields of scalars or vectors.
        ymax: `float`. Upper bound in the Y dimension for array fields of scalars or vectors.
        zmin: `float`. Lower bound in the Z dimension for array fields of scalars or vectors.
        zmax: `float`. Upper bound in the Z dimension for array fields of scalars or vectors.
        bounds: `array_like`. Array of bounds (bounding box or rectangle) for array fields of scalars or vectors.
            This array can be of size 6 (xmin, xmax, ymin, ymax, zmin, zmax) for 3D fields,
            or of size 4 (xmin, xmax, ymin, ymax) for 2D fields.
            Takes precedence over separate xmin, xmax, ... etc. arguments.
        translation=kwargs.get('translation'),
        rotation=kwargs.get('rotation'),
        scaling=kwargs.get('scaling'),
        model_matrix
    """

    if 'transform' in kwargs:
        transform = kwargs['transform']
        if not isinstance(transform, Transform):
            raise ValueError('Provided transform argument is not a Transform object')
    else:
        separate_bounds = [
            kwargs.get('xmin', -0.5),
            kwargs.get('xmax', 0.5),
            kwargs.get('ymin', -0.5),
            kwargs.get('ymax', 0.5),
            kwargs.get('zmin', -0.5),
            kwargs.get('zmax', 0.5)
        ]

        transform_kwargs = dict(
            bounds=kwargs.get('bounds', separate_bounds),
            translation=kwargs.get('translation'),
            rotation=kwargs.get('rotation'),
            scaling=kwargs.get('scaling'),
            custom_matrix=kwargs.get('model_matrix')
        )
        transform = Transform(**transform_kwargs)

    transform.add_drawable(drawable)
    transform.parent_updated()
    drawable.transform = transform

    return drawable



from __future__ import print_function
import warnings
import types
import codecs

import ipywidgets as widgets
from traitlets import Unicode, Bool, Int, List

from ._version import __version__
from .objects import Drawable, ListOrArray


class Plot(widgets.DOMWidget):
    """
    Main K3D widget.

    Attributes:
        antialias: `bool`: Enable antialiasing in WebGL renderer, changes have no effect after displaying.
        height: `int`: Height of the Widget in pixels, changes have no effect after displaying.
        background_color: `int`.  Packed RGB color of the plot background (0xff0000 is red, 0xff is blue).
        camera_auto_fit: `bool`. Enable automatic camera setting after adding, removing or changing a plot object.
        grid_auto_fit: `bool`. Enable automatic adjustment of the plot grid to contained objects.
        voxel_paint_color: `int`. The (initial) int value to be inserted when editing voxels.
        grid: `array_like`. 6-element tuple specifying the bounds of the plot grid (x0, y0, z0, x1, y1, z1).
        camera: `array_like`. 9-element list or array specifying camera position.
        objects: `list`. List of `k3d.objects.Drawable` currently included in the plot, not to be changed directly.
    """

    _view_name = Unicode('PlotView').tag(sync=True)
    _model_name = Unicode('PlotModel').tag(sync=True)
    _view_module = Unicode('k3d').tag(sync=True)
    _model_module = Unicode('k3d').tag(sync=True)

    _view_module_version = Unicode('~' + __version__).tag(sync=True)
    _model_module_version = Unicode('~' + __version__).tag(sync=True)

    # readonly (specified at creation)
    antialias = Bool().tag(sync=True)
    height = Int().tag(sync=True)

    # readonly (not to be modified directly)
    object_ids = List().tag(sync=True)

    # read-write
    camera_auto_fit = Bool(True).tag(sync=True)
    grid_auto_fit = Bool(True).tag(sync=True)
    grid = ListOrArray((-1, -1, -1, 1, 1, 1), minlen=6, maxlen=6).tag(sync=True)
    background_color = Int().tag(sync=True)
    voxel_paint_color = Int().tag(sync=True)
    camera = ListOrArray(minlen=9, maxlen=9, empty_ok=True).tag(sync=True)
    screenshot = Unicode().tag(sync=True)

    objects = []

    def __init__(self, antialias=True, background_color=0xFFFFFF, camera_auto_fit=True, grid_auto_fit=True, height=512,
                 voxel_paint_color=0, grid=(-1, -1, -1, 1, 1, 1), *args, **kwargs):
        super(Plot, self).__init__()

        self.antialias = antialias
        self.camera_auto_fit = camera_auto_fit
        self.grid_auto_fit = grid_auto_fit
        self.grid = grid
        self.background_color = background_color
        self.voxel_paint_color = voxel_paint_color
        self.height = height

        self.object_ids = []
        self.objects = []

        self._screenshot_handler = None
        self.observe(self._screenshot_changed, names=['screenshot'])

    def __iadd__(self, objs):
        assert isinstance(objs, Drawable)

        for obj in objs:
            if obj.id not in self.object_ids:
                self.object_ids = self.object_ids + [obj.id]
                self.objects.append(obj)

        return self

    def __add__(self, objs):
        warnings.warn('Using plus operator to add objects to plot is discouraged in favor of +=')
        return self.__iadd__(objs)

    def __isub__(self, objs):
        assert isinstance(objs, Drawable)

        for obj in objs:
            self.object_ids = [id_ for id_ in self.object_ids if id_ != obj.id]
            if obj in self.objects:
                self.objects.remove(obj)

        return self

    def __sub__(self, objs):
        warnings.warn('Using minus operator to remove objects from plot is discouraged in favor of -=')
        return self.__isub__(objs)

    def display(self, **kwargs):
        super(Plot, self)._ipython_display_(**kwargs)

    def fetch_screenshot(self, handler=None):
        self._screenshot_handler = handler
        if isinstance(self._screenshot_handler, types.GeneratorType):
            if handler is not self._screenshot_handler:
                # start (only new) generator
                next(self._screenshot_handler)
        self.send({'msg_type': 'fetch_screenshot'})

    def _screenshot_changed(self, change):
        if self._screenshot_handler is not None:
            data = codecs.decode(change['new'].encode('ascii'), 'base64')
            if isinstance(self._screenshot_handler, types.GeneratorType):
                try:
                    self._screenshot_handler.send(data)
                except StopIteration:
                    # unregister used up generator
                    self._screenshot_handler = None
            elif callable(self._screenshot_handler):
                self._screenshot_handler(data)
            else:
                raise TypeError('Screenshot handler of wrong type')
